package it.unina.tools.csvreport;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Arrays;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;

import it.unina.tools.loadandsave.FileLoadAndSaveUtils;
import it.unina.tools.loadandsave.LoadAndSaveConfiguration;

/**
 * Classe per recuperare i risultati di un test a partire da un
 * report .csv
 * @author francesco
 *
 */
public class CsvParser {
	
	public static final String QUESTIONMARKID = "?-Operator Coverage";
	public static final String BRANCHID = "Branch Coverage";
	public static final String LOOPID = "Loop Coverage";
	public static final String STATEMENTID = "Statement Coverage";
	public static final String SYNCHRONIZEDID = "Synchronized Coverage";
	public static final String TERMID = "Term Coverage";
	
	
	
	public static final String PACKAGEID = "package";
	public static final String CLASSID = "class";
	public static final String METHODID = "method";
	
	private byte[] csvFile;
	private List<CoverageResult> coverageResults;
	
	/**
	 * in previsione dell'utilizzo su cloud abbiamo inserito un costruttore
	 * che prende in input un byte[]
	 * @param csvFile byte[] contenente il report csv
	 */
	public CsvParser(byte[] csvFile){
		this.csvFile = csvFile;
		try{
			extractResult();
		} catch (CsvReaderException e){
			this.coverageResults=null;
			e.printStackTrace();
		}
	}
	
	/**
	 * per consentire un utilizzo stand-alone c'è anche un costruttore
	 * che prende in input il nome del file e utilizza le librerie
	 * it.unina.loadandsave per ottenere un byte[]
	 * @param csvFileName nome del file contenente il report csv
	 */
	public CsvParser(String csvFileName){
		this(new FileLoadAndSaveUtils().load(new LoadAndSaveConfiguration(csvFileName)));
	}
	
	/**
	 * metodo che opera sul byte[] e recupera i risultati contenuti
	 * @throws CsvReaderException
	 */
	private void extractResult() throws CsvReaderException{
		//estraiamo i risultati dal file csv
		 
		List<CoverageResult> results = new LinkedList<CoverageResult>();
		try{
			BufferedReader br = new BufferedReader(new InputStreamReader(new ByteArrayInputStream(this.csvFile)));
			String strLine = "";
			Iterator<String> st = null;
			if((strLine = br.readLine())==null) throw new CsvReaderException();
			
			//dalla prima riga estraiamo il nome dei criteri calcolati
			st=Arrays.asList(processLine(strLine).split(",")).iterator(); //recuperiamo la riga
			st.next(); //ignoriamo l'elemento package
			st.next(); //ignoriamo l'elemento item
			st.next(); //ignoriamo l'elemento type
			
			List<String> criteriaName = new LinkedList<String>();
			while (st.hasNext()){
				criteriaName.add(st.next().replaceAll(" covered", "")); //leggiamo il nome del criterio
				st.next(); //ignoriamo l'elemento uncovered
				st.next(); //ignoriamo l'elemento total
				st.next(); //ignoriamo l'elemento coverage
				st.next(); //ignoriamo l'elemento coverage %
			}
			if(criteriaName.isEmpty()) throw new CsvReaderException();
			
			//leggiamo dalla seconda in poi i risultati
			while( (strLine = br.readLine()) != null){
				//iteratore sugli elementi della riga separati da ","
				st=Arrays.asList(processLine(strLine).split(",")).iterator(); 
				String packageName = st.next(); //leggiamo il package
				String item = st.next(); //leggiamo l'item
				String type = st.next(); //leggiamo il type
				
				List <CoverageCriterion> criteria = new LinkedList<CoverageCriterion>();
				
				//recuperiamo i risultati per ogni criterio presente
				for(String criterionName:criteriaName){
					Integer covered = Integer.parseInt(st.next()); //leggiamo l'elemento covered
					Integer uncovered = Integer.parseInt(st.next()); //leggiamo l'elemento uncovered
					Integer total = Integer.parseInt(st.next()); //leggiamo l'elemento total
					Double coverage = Double.parseDouble(st.next()); //leggiamo l'elemento coverage
					st.next(); //ignoriamo l'elemento coverage % poichè è ridondante
					
					criteria.add(new CoverageCriterion(criterionName, covered, uncovered, total, coverage));
				}
				results.add(new CoverageResult(packageName, item, type, criteria));
				
			}

		} catch (NoSuchElementException e){
			throw new CsvReaderException();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		if (results.isEmpty()) throw new CsvReaderException();
		
		this.coverageResults = results;
	}
	
	/**
	 * metodo che processa un rigo per la successiva fase di split su ","
	 * poichè nel file csv sono presenti valori contenenti , che non devono
	 * essere splittati (racchiusi tra virgolette) processiamo la riga
	 * eliminando tali virgolette e sostituendo i caratteri "," che non
	 * rappresentano separatori con caratteri "."
	 * @param strLine stringa da processare
	 * @return la stringa processata
	 */
	public static String processLine(String strLine){
		/*
		 * inseriamo dei caratteri speciali "&&&" all'inizio e alla fine della
		 * stringa da processare. Facciamo questo per poter trattare il problema 
		 * genericamente. Infatti se nella riga c'è una (o più) porzione racchiusa
		 * tra " splittando su "\"" otteniamo delle porzioni separate ma non
		 * sappiamo quali di queste erano quelle esterne e quali erano quelle 
		 * interne alle virgolette. Così facendo invece le porzioni che erano
		 * tra virgolette sono sempre gli elementi con indice pari del risultato
		 * dello split
		 */
		strLine="&&&"+strLine+"&&&";
		String[] s = strLine.split("\"");

		String temp = null;
		List<String> ris = new LinkedList<String>();
		
		// analizziamo il risultato dello split
		if (s.length>1){
			/*
			 * se c'è più di una porzione allora erano presenti virgolette, e
			 * quindi ci sono porzioni da analizzare
			 */
			
			int index = 1;
			for (String st:s){
				if(index % 2 == 0){
					/*
					 * gli elementi pari sono quelli che erano racchiusi tra
					 * virgolette; sostituiamo le eventuali "," con "."
					 */
					st=st.replaceAll(",", ".");
				}
				ris.add(st);
				index++;
			}
			
			// ricostruiamo un'unica stringa con i frammenti processati
			temp="";
			for(String st : ris){
				temp+=st;
			}
		}else{
			//se non vi erano virgolette allora non bisogna fare processi
			temp=s[0];
		}
		
		//rimuoviamo i caratteri speciali "&&&"
		strLine=temp.replaceAll("&&&", "");
		
		return strLine;
	}

	public List<CoverageResult> getCoverageResults() {
		return coverageResults;
	}
	
	/**
	 * restituisce una stringa contenente i risultati come stringa
	 * @return la stringa con i risultati
	 */
	public String getCoverageResultsAsString() {
		String res = "";
		int i=1;
		for (CoverageResult r:coverageResults){
			res+="Risultato " + i + 
					"\npackage:\t"+r.getPackageName() +
					"\nitem:\t"+r.getItem()+
					"\ntype:\t"+r.getType();
			for (CoverageCriterion crit : r.getCoverageResults()){
				res+="\n"+crit.getName()+
						"\ncovert: "+crit.getCovered()+
						"\tuncovert: "+crit.getUncovered()+
						"\ttotal: "+crit.getTotal()+
						"\tcoverage: "+crit.getCoverage();
			}
			i++;
			res+="\n\n";
		}
		return res;
	}
	
	/**
	 * Metodo che dato il nome di un metodo e un criterio di coverage restituisce
	 * la coverage calcolata per quel criterio
	 * @param method metodo ci cui si vuole sapere la coverage
	 * @param coverageCriterion criterio di coverage desiderato. DEVE essere 
	 * una delle stringhe definite in CsvParser
	 * @return
	 */
	public Double getCoverage(String method, String coverageCriterion){
		if (this.coverageResults == null) return -4d; //RISULTATI NON PRESENTI
		for (CoverageResult r : this.coverageResults){
			if (r.getType().equals(CsvParser.METHODID)){
				String[] packageSplitted = r.getPackageName().split("\\.");
				String itemName = packageSplitted[packageSplitted.length-1];
				if (itemName.equals(method)){
					for (CoverageCriterion crit : r.getCoverageResults()){
						if(crit.getName().equals(coverageCriterion)){
							if(crit.getTotal()==0) return -1d; //COVERAGE = 1 POICHE' NON C'ERANO ELEMENTI DA COPRIRE
							return crit.getCoverage();
						}	
					}
					return -2d; //CRITERIO NON TROVATO
				}
			}
		}
		return 0d; //METODO NON TROVATO
	}
}
